
require "Window"
require "Sound"
require "Spell"
require "Apollo"
require "GameLib"
require "AbilityBook"
require "MacrosLib"

-----------------------------------------------------------------------------------------------
-- AMPBuilds Module Definition
-----------------------------------------------------------------------------------------------
local AMPBuilds = {}

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------

local karTabTypes =
{
  Builds = 1,
  Stats = 2,
  Options = 3
}

--[[
"DDBagItem" - Any item
"DDSpellbookItem" - Abilities like the ones in the Action Set Builder
"DDMacro" - Macros
"DDNonCombat" - Non-Combat spells, such as Pets, Mounts, or Recall spells
"DDGameCommand" - Commands like the ones found in the Command tab in the Non-Combat Spellbook
]]

local karTabDragDropType =
{
  [karTabTypes.Builds] = "DDMacro", --"DDGameCommand"
  [karTabTypes.Stats] = "Stats", --"DDNonCombat",
  [karTabTypes.Options] = "Options" --"DDNonCombat",
}

local knCategoryUtilityId 			= 1
local knCategorySupportId 			= 2
local knCategoryDamageId 			= 3
local knCategoryDamageSupportId 	= 4
local knCategoryDamageUtilitytId 	= 5
local knCategorySupportUtilityId 	= 6

local karCategoryToConstantData =
{
	[knCategoryUtilityId] 			= {strLightBulbName = "LightBulbS",		strLightBulbSprite = "spr_AMPS_MiddleGlow_S", 	strName = Apollo.GetString("AMP_Utility"),		strLabelName = "LabelUtility",		strButtonLabel = "UtilityLabel" },
	[knCategorySupportId]			= {strLightBulbName = "LightBulbNE",	strLightBulbSprite = "spr_AMPS_MiddleGlow_NE",	strName = Apollo.GetString("AMP_Support"), 		strLabelName = "LabelSupport",		strButtonLabel = "SupportLabel" },
	[knCategoryDamageId] 			= {strLightBulbName = "LightBulbNW",	strLightBulbSprite = "spr_AMPS_MiddleGlow_NW", 	strName = Apollo.GetString("AMP_Assault"),		strLabelName = "LabelAssault",		strButtonLabel = "AssaultLabel" },
	[knCategoryDamageSupportId] 	= {strLightBulbName = "LightBulbN",		strLightBulbSprite = "spr_AMPS_MiddleGlow_N", 	strName = Apollo.GetString("AMP_Hybrid"),		strLabelName = "LabelHybrid",		strButtonLabel = "AssaultSupportLabel" },
	[knCategoryDamageUtilitytId] 	= {strLightBulbName = "LightBulbSE",	strLightBulbSprite = "spr_AMPS_MiddleGlow_SW", 	strName = Apollo.GetString("AMP_PvPOffense"),	strLabelName = "LabelPvPOffense",	strButtonLabel = "AssaultUtilityLabel" },
	[knCategorySupportUtilityId] 	= {strLightBulbName = "LightBulbSW",	strLightBulbSprite = "spr_AMPS_MiddleGlow_SE", 	strName = Apollo.GetString("AMP_PvPDefense"),	strLabelName = "LabelPvPDefense",	strButtonLabel = "UtilitySupportLabel" },
}

local karCategoriesInClockwiseOrder =
{
	knCategoryUtilityId,
	knCategoryDamageUtilitytId,
	knCategoryDamageId,
	knCategoryDamageSupportId,
	knCategorySupportId,
	knCategorySupportUtilityId,
}

local knDefaultZoom						= 21.31				-- 21.47 without top banner
local knLowestZoom						= 21.31				-- 21.47 without top banner
local knColumnsPerCategory 				= 8 				-- How many columns exist for each category (columns)
local knNumRows 						= 12 				-- Number of rows in the radial grid

local knAugmentationActivatedColor 		= "btn_AMPS_CirclePressed"		-- On
local knAugmentationInactivatedColor 	= "btn_AMPS_CircleNormal"		-- Can Click
local knAugmentationAlmostThereColor 	= "spr_AMPs_LockStretch_Blue"	-- 1 Away from Can Click
local knAugmentationInaccessibleColor 	= "spr_AMPs_LockStretch_Grey"
local knAugmentationUnavailableColor 	= "spr_AMPs_LockStretch_Red"

local ktAugmentationValidationResult =
{
	[ActionSetLib.CodeEnumLimitedActionSetResult.Ok]									= Apollo.GetString("EldanAugmentation_Ok"),
	[ActionSetLib.CodeEnumLimitedActionSetResult.EldanAugmentation_LockedInlaid]		= Apollo.GetString("EldanAugmentation_LockedInlaidAugmentation"),
	[ActionSetLib.CodeEnumLimitedActionSetResult.EldanAugmentation_LockedCategoryTier]	= Apollo.GetString("EldanAugmentation_LockedCategoryTier"), -- TODO rename to rank
	[ActionSetLib.CodeEnumLimitedActionSetResult.UnknownClassId]						= Apollo.GetString("EldanAugmentation_InvalidClass"),
	[ActionSetLib.CodeEnumLimitedActionSetResult.EldanAugmentation_InvalidSeries] 		= Apollo.GetString("EldanAugmentation_InvalidSeries"),
	[ActionSetLib.CodeEnumLimitedActionSetResult.EldanAugmentation_InvalidId]			= Apollo.GetString("EldanAugmentation_InvalidEldanAugmentationId"),
	[ActionSetLib.CodeEnumLimitedActionSetResult.EldanAugmentation_InvalidCategoryId]	= Apollo.GetString("EldanAugmentation_InvalidAugmentationCategoryId"),
	[ActionSetLib.CodeEnumLimitedActionSetResult.EldanAugmentation_InvalidCategoryTier]	= Apollo.GetString("EldanAugmentation_InvalidAugmentationCategoryTier"),
}


-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function AMPBuilds:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self

  -- Set Default settings
  o.config = {  MainPosition = { nLeft = 20, nTop = 20, nRight = 444, nBottom = 620 },
                ConfirmDlgPosition = { nLeft = 20, nTop = 20, nRight = 338, nBottom = 293 }, }
  o.nSelectedTab = karTabTypes.Builds
  o.savedAMPBuildList = {}
  o.statsList = {}
  o.optionsList = { optOut = false, }
	o.tWndRefs = {}

  return o
end

function AMPBuilds:Init()
  local bHasConfigureFunction = true
  local strConfigureButtonText = "AMP Builds"
  local tDependencies = { "AbilityAMPs",
                          "Macros", }

  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

function AMPBuilds:OnDependencyError(strDependency, strError)
  return true
end


-----------------------------------------------------------------------------------------------
-- AMPBuilds OnLoad
-----------------------------------------------------------------------------------------------
function AMPBuilds:OnLoad()
  self.xmlDoc = XmlDoc.CreateFromFile("AMPBuilds.xml")
  self.xmlDoc:RegisterCallback("OnDocumentReady", self) 
end

function AMPBuilds:OnDocumentReady()
  if self.xmlDoc == nil then
    return
  end
  
  Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded",   "OnInterfaceMenuListHasLoaded", self)
  Apollo.RegisterEventHandler("WindowManagementReady",     "OnWindowManagementReady", self)
  Apollo.RegisterEventHandler("CharacterEldanAugmentationsUpdated", "OnCharacterEldanAugmentationsUpdated", self)
	-- Apollo.RegisterEventHandler("PlayerCurrencyChanged", "OnPlayerCurrencyChanged", self)

  
  --Apollo.RegisterEventHandler("OpenAMPBuilds", "OnAMPBuildsOn", self)
  Apollo.RegisterEventHandler("ToggleAMPBuilds", "OnToggleAMPBuilds", self)
  Apollo.RegisterSlashCommand("ampbuilds", "OnToggleAMPBuilds", self)
  Apollo.RegisterTimerHandler("AMPBuildLoadIconTimer", "OnLoadAMPBuildIcons", self)
  
  -- load our forms
  self.wndMain = Apollo.LoadForm(self.xmlDoc, "AMPBuildsForm", nil, self)
  self.wndMain:SetAnchorOffsets(self.config.MainPosition.nLeft, self.config.MainPosition.nTop, self.config.MainPosition.nRight, self.config.MainPosition.nBottom)
  self.wndMain:Show(false)
  
  self.wndEntryContainer = self.wndMain:FindChild("EntriesContainer")
  self.wndEntryContainerStats = self.wndMain:FindChild("EntriesContainerStats")
  self.wndEntryContainerMisc = self.wndMain:FindChild("EntriesContainerMisc")
  self.wndTabsContainer = self.wndMain:FindChild("TabsContainer")
  self.wndTabsContainer:FindChild("BankTabBtnCmd"):SetData(karTabTypes.Builds)
  self.wndTabsContainer:FindChild("TabBtnStats"):SetData(karTabTypes.Stats)
  self.wndTabsContainer:FindChild("BankTabBtnMisc"):SetData(karTabTypes.Options)

  self.editBox = self.wndMain:FindChild("EditBox")
  self.curIconSelection = self.wndMain:FindChild("curIconSelection")

  self.selectedItem = nil
  self.wndSelectedIcon = nil
  self.wndAMPsDisplay = nil

	self.tWndRefs.wndMain = nil
	self.nVScrollPos = 170
	self.nZoomLevel = knDefaultZoom

  if self.optionsList.optOut == nil then self.optionsList.optOut = false end

  self.nAmpRespecUpdates = 0
  self.buildSwitched = false
  self.buildData = nil
  -- self.nPlayerCurrency = 0
  -- self.nActualRespecCost = 0
end

function AMPBuilds:OnWindowManagementReady()
  Event_FireGenericEvent("WindowManagementAdd", {wnd = self.wndMain, strName = "AMPBuilds"})
end

function AMPBuilds:OnInterfaceMenuListHasLoaded()
  Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "AMP Builds", {"ToggleAMPBuilds", "", "Icon_Windows32_UI_CRB_InterfaceMenu_NonCombatAbility"})
end

-- function AMPBuilds:OnPlayerCurrencyChanged()
-- Unreliable.  Currency updates fire at indeterminate intervals.  Several updates fire after the amp switch has completed...
--[[
  if self.buildSwitched then
    local nUpdatedCurrency = GameLib.GetPlayerCurrency():GetAmount()
    local nAmountDeducted = self.nPlayerCurrency - nUpdatedCurrency
    self.nPlayerCurrency = nUpdatedCurrency
    Print("Actual amount deducted: " .. tostring(nAmountDeducted))
    self.nActualRespecCost = self.nActualRespecCost + nAmountDeducted
  end
]]
-- end

function AMPBuilds:OnConfigure()
  self.nSelectedTab = karTabTypes.Options
  self:OnAMPBuildsOn()
end

function AMPBuilds:SaveWindowPos()
  self.config.MainPosition.nLeft, self.config.MainPosition.nTop, self.config.MainPosition.nRight, self.config.MainPosition.nBottom = self.wndMain:GetAnchorOffsets()
  if self.dlgConfirm ~= nil then
    self.config.ConfirmDlgPosition.nLeft, self.config.ConfirmDlgPosition.nTop, self.config.ConfirmDlgPosition.nRight, self.config.ConfirmDlgPosition.nBottom = self.dlgConfirm:GetAnchorOffsets()
  end
end

function AMPBuilds:OnSave(eLevel)
  if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
    return nil
  end

  self:SaveWindowPos()
  local tSaveData = { tAMPBuildList = deepcopy(self.savedAMPBuildList),
                      tOptionsList = deepcopy(self.optionsList),
                      tConfig = deepcopy(self.config), }

  return tSaveData
end

function AMPBuilds:OnRestore(eLevel, tData)
  if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
    return
  end

  if tData == nil then
    return
  end

  if tData.tAMPBuildList ~= nil then
    self.savedAMPBuildList = tData.tAMPBuildList
  end
  if tData.tOptionsList ~= nil then
    self.optionsList = tData.tOptionsList
  end
  if tData.tConfig ~= nil then
    self.config = tData.tConfig
  end

  if self.savedAMPBuildList == nil then
    self.savedAMPBuildList = {}
  end
  if self.optionsList == nil then
    self.optionsList = {}
  end
  if self.config == nil then
    self.config = {}
  end
end

function AMPBuilds:OnToggleAMPBuilds()
  if self.wndMain and self.wndMain:IsValid() and self.wndMain:IsVisible() then
    self.wndMain:Close()
  else
    self:OnAMPBuildsOn()
  end
end

function AMPBuilds:OnAMPBuildsOn()
  self.nSelectedTab = self.nSelectedTab or karTabTypes.Builds
  self:Redraw()
  self.wndMain:Show(true)
  self.wndMain:ToFront()
end

function AMPBuilds:Redraw()
  local unitPlayer = GameLib.GetPlayerUnit()  
  if not unitPlayer then
    return
  end
  
  self:ShowTab()
end

function AMPBuilds:ShowTab()
  self.wndEntryContainer:DestroyChildren()
  self.wndEntryContainerStats:DestroyChildren()
  self.wndEntryContainerMisc:DestroyChildren()
      
  self.wndEntryContainer:Show(self.nSelectedTab == karTabTypes.Builds)
  self.wndEntryContainerMisc:Show(self.nSelectedTab == karTabTypes.Options)
  self.wndEntryContainerStats:Show(self.nSelectedTab == karTabTypes.Stats)
  self.selectedItem = nil

  local function SortFunction(a,b)
    local aData = a and a:GetData()
    local bData = b and b:GetData()
    if not aData and not bData then
      return true
    end
    return (aData.strName or aData.strName) < (bData.strName or bData.strName)
  end

  if self.nSelectedTab == karTabTypes.Builds then
    for idx, tData in pairs(self.savedAMPBuildList) do
      self:HelperCreateAMPBuildEntry(tData)
    end
    self.wndEntryContainer:ArrangeChildrenVert(0, 1)
    self.wndEntryContainer:SetText(#self.wndEntryContainer:GetChildren() == 0 and "No AMP Builds saved yet." or "")
    self:ShowBuildItems(true)

  elseif self.nSelectedTab == karTabTypes.Stats then
    for idx, tData in pairs(self.statsList) do
      self:HelperCreateStatsItem(tData)
    end
    self.wndEntryContainerStats:ArrangeChildrenVert(0, 1)
    self.wndEntryContainerStats:SetText(#self.wndEntryContainerStats:GetChildren() == 0 and "No Stats obtained yet." or "")
    self:ShowBuildItems(false)

  elseif self.nSelectedTab == karTabTypes.Options then
    self:HelperCreateOptions()
    self.wndEntryContainerMisc:ArrangeChildrenTiles()
    self:ShowBuildItems(false)
  end
  
  for idx, wndTab in pairs(self.wndTabsContainer:GetChildren()) do
    wndTab:SetCheck(self.nSelectedTab == wndTab:GetData())
  end
end

function AMPBuilds:ShowBuildItems(bShow)
  local tFooterItems = self.wndMain:FindChild("Footer"):GetChildren()
  for idx, child in pairs(tFooterItems) do
    child:Show(bShow or false)
  end
end

function AMPBuilds:HelperCreateStatsItem(tData)
  local wndEntry = Apollo.LoadForm(self.xmlDoc, "StatsListItem", self.wndEntryContainerStats, self)
  wndEntry:SetData(tData)
  wndEntry:FindChild("Info"):SetText(tData.strInfo)
  wndEntry:FindChild("Icon"):SetSprite(tData.strSprite)
  wndEntry:FindChild("CashWindow"):SetAmount(tData.nRespecAmt)
end

function AMPBuilds:HelperCreateOptions()
  local wndEntry = Apollo.LoadForm(self.xmlDoc, "OptionsForm", self.wndEntryContainerMisc, self)
  wndEntry:FindChild("OptOut"):FindChild("CheckBox"):SetCheck(self.optionsList.optOut or false)
end

function AMPBuilds:HelperCreateAMPBuildEntry(tData)
  local wndEntry = Apollo.LoadForm(self.xmlDoc, "BuildListItem", self.wndEntryContainer, self)  
  wndEntry:SetData(tData)
  wndEntry:FindChild("Title"):SetText(tData.strName)
  wndEntry:FindChild("Icon"):SetSprite(tData.strSprite)
end

function AMPBuilds:HelperCreateMessage(strMessage)
  self:AMPPrint(strMessage)
end

function AMPBuilds:OnClose()
  self.wndMain:Show(false)
end

function AMPBuilds:OnCloseAMPsDisplay()
  if self.wndAMPsDisplay then
    self.wndAMPsDisplay:Show(false)
  end
end

function AMPBuilds:OnTabBtnCheck(wndHandler, wndControl, eMouseButton)
  self.nSelectedTab = wndControl:GetData()
  self:ShowTab()
end

function AMPBuilds:OnOptOutCheck(wndHandler, wndControl)
  self.optionsList.optOut = true
end

function AMPBuilds:OnOptOutUncheck(wndHandler, wndControl)
  self.optionsList.optOut = false
end

function AMPBuilds:OnMouseEnter()
  self.wndMain:FindChild("Header"):SetFont("CRB_Interface16_O")
end

function AMPBuilds:OnMouseExit()
  self.wndMain:FindChild("Header"):SetFont("CRB_AlienHuge")
end

function AMPBuilds:OnListItemCheck(wndHandler, wndControl, eMouseButton, nPosX, nPosY, bDoubleClick)
  if eMouseButton == GameLib.CodeEnumInputMouse.Right then
    local selectedBuild = wndControl:GetParent()
    -- local strBuildName = selectedBuild:FindChild("Title"):GetText()
    -- Print(tostring(wndControl:GetName()))
    -- Print(strBuildName)
    local tData = selectedBuild:GetData()
    -- Print(tostring(tData))

    self:DrawAMPBuild(tData)
    --[[
    if not self.wndAMPsDisplay then
      self.wndAMPsDisplay = Apollo.LoadForm(self.xmlDoc, "AMPsDisplayForm", nil, self)
    end
    self.wndAMPsDisplay:Show(true)
    ]]
  else
    if self.selectedItem == nil then
      self.selectedItem = wndControl:GetParent()
      self.selectedItem:FindChild("Window"):Show(true)
    else
      self.selectedItem:FindChild("Window"):Show(false)
      if self.selectedItem ~= wndControl:GetParent() then
        self.selectedItem = wndControl:GetParent()
        self.selectedItem:FindChild("Window"):Show(true)
      else
        self.selectedItem = nil
      end
    end
  end
end

function AMPBuilds:OnListItemUncheck(wndHandler, wndControl)
  self.selectedItem:FindChild("Window"):Show(false)
  self.selectedItem = nil
end

function AMPBuilds:OnAddNewBuild(wndHandler, wndControl)
  local name = self.editBox:GetText()
  if name == "" or name == nil then
    self:AMPPrint("Enter a name in the text box.")
    return
  end

  local exists = self.savedAMPBuildList[name]
  if exists then
    self:AMPPrint("This build already exists.")
    return
  end

  local curIcon = self.curIconSelection:GetSprite()
  if curIcon == "sprAS_Prompt_Resource2" then
    curIcon = "UI_BK3_StoryPanelAlert_Icon"
  end

  local curSpec = AbilityBook.GetCurrentSpec()
  local tAugs = self:GetCurrentEldanAugmentationData()
  local nAugCount = tAugs.nCount

  local id = self:CreateMacro(curIcon, name)

  local tData = { strName = name,
      nSpec = curSpec,
      strSprite = curIcon,
      tUnlockedAugments = tAugs,
      nMacroId = id,
      nCount = nAugCount }

  self.savedAMPBuildList[name] = tData

  self:Redraw()
  self.curIconSelection:SetSprite("CRB_ActionBarIconSprites:sprAS_Prompt_Resource2")
  self.editBox:SetText("")
  self:AMPPrint("Created a new AMP Build: " .. name)
end

function AMPBuilds:GetCurrentEldanAugmentationData()
  local tEldanAugmentationData = AbilityBook.GetEldanAugmentationData(AbilityBook.GetCurrentSpec())
  if not tEldanAugmentationData then
    self:AMPPrint("Failed to GetEldanAugmentationData")
    return nil
  end
  
  -- Create a map of dependencies and amps for easy access
  local count = 0
  local tUnlockedAugments = { nCount = 0,
                              tAmps = { [1] = {}, [2] = {}, [3] = {} }, }

  for idx, tAmp in pairs(tEldanAugmentationData.tAugments) do
    if tAmp.eEldanAvailability == AbilityBook.CodeEnumEldanAvailability.Activated then
      count = count + 1

      local tAug = {}
      tAug.nId = tAmp.nId
      tAug.strTitle = tAmp.strTitle
      tAug.nCategoryId = tAmp.nCategoryId
      tAug.nCategoryTier = tAmp.nCategoryTier
      tAug.nDisplayRow = tAmp.nDisplayRow
      tAug.nDisplayColumn = tAmp.nDisplayColumn
      tAug.nPowerCost = tAmp.nPowerCost
      
      tUnlockedAugments.nCount = tUnlockedAugments.nCount + 1
      table.insert(tUnlockedAugments.tAmps[tAmp.nCategoryTier], tAug)
    end
  end

  return tUnlockedAugments
  --return tEldanAugmentationData
end

function AMPBuilds:GetCurrentSpecUnlockedAugments()
  local tEldanAugmentationData = AbilityBook.GetEldanAugmentationData(AbilityBook.GetCurrentSpec())
  if not tEldanAugmentationData then
    self:AMPPrint("AMPBuilds:GetCurrentSpecUnlockedAugments() - Failed to GetEldanAugmentationData")
    return nil
  end
  
  local tUnlockedAugments = { nCount = 0,
                              tAmps = { [1] = {}, [2] = {}, [3] = {} }, }

  for idx, tAmp in pairs(tEldanAugmentationData.tAugments) do
    if tAmp.eEldanAvailability == AbilityBook.CodeEnumEldanAvailability.Activated then
      tUnlockedAugments.nCount = tUnlockedAugments.nCount + 1
      table.insert(tUnlockedAugments.tAmps[tAmp.nCategoryTier], { nId = tAmp.nId })
    end
  end

  return tUnlockedAugments
end

function AMPBuilds:GetUnlockedAugments(tData)
  if not tData then
    self:AMPPrint("AMPBuilds:GetUnlockedAugments() - Failed to supply valid table")
    return nil
  end
  
  local tUnlockedAugments = {}

  for idx, tAmp in pairs(tData.tAugments) do
    if tAmp.eEldanAvailability == AbilityBook.CodeEnumEldanAvailability.Activated then
      tUnlockedAugments[tAmp.nId] = tAmp.nId
    end
  end

  return tUnlockedAugments
end

function AMPBuilds:OnUpdateBuild()
  if self.selectedItem == nil then
    self:AMPPrint("Select an item from the list to update.")
    return
  end

  local buildData = self.selectedItem:GetData()
  if buildData == nil then
    self:AMPPrint("Problem retrieving build data.")
    return
  end

  local name = self.editBox:GetText()
  if name ~= "" then
    self.savedAMPBuildList[buildData.strName] = nil
    buildData.strName = name
    self.savedAMPBuildList[buildData.strName] = buildData
  end

  local curSprite = self.curIconSelection:GetSprite()
  if curSprite ~= "sprAS_Prompt_Resource2" then
    self.savedAMPBuildList[buildData.strName].strSprite = curSprite
  end

  local tParam = {  nId = buildData.nMacroId,
                    bGlobal = false,
                    sName = buildData.strName,
                    sSprite = buildData.strSprite,
                    sCmds = "/eval CallAMPBuildsMacro(\"" .. buildData.strName .. "\")", } 

  self:SaveMacro(tParam)
  self:Redraw()
  self.curIconSelection:SetSprite("CRB_ActionBarIconSprites:sprAS_Prompt_Resource2")
  self.editBox:SetText("")
  self:AMPPrint("Updated build: " .. buildData.strName)

end

function AMPBuilds:OnDeleteBuild()
  if self.selectedItem == nil then
    self:AMPPrint("Select an item from the list to delete.")
    return
  end

  local buildData = self.selectedItem:GetData()
  if buildData == nil then
    self:AMPPrint("Problem retrieving build data.")
    return
  end

  self.savedAMPBuildList[buildData.strName] = nil
  self:DeleteMacro(buildData.nMacroId)
  self:Redraw()
  self:AMPPrint("Deleted build: " .. buildData.strName)

end

function AMPBuilds:OpenIconWnd()
  if self.wndIcon == nil then
    self.wndIcon = Apollo.LoadForm(self.xmlDoc, "SelectIcon", nil, self)
    self.bIconsLoaded = false;
    Apollo.CreateTimer("AMPBuildLoadIconTimer", 0.1, false)
    Apollo.StartTimer("AMPBuildLoadIconTimer")
  else
    self.wndIcon:ToFront()
  end
end

function AMPBuilds:OnLoadAMPBuildIcons()
  self.wndIconList = self.wndIcon:FindChild("IconList")
  self.wndIconList:SetFocus()
  
  -- create the list of icons
  local arStrMacroIcons = MacrosLib.GetMacroIconList()
   
  local wndFirstIcon = nil;
  for idx = 1, #arStrMacroIcons do  
    if self.wndIcon == nil then -- in case when user quits before done loading
      break
    end
    
    local wnd = Apollo.LoadForm(self.xmlDoc, "IconItem", self.wndIconList, self)
    if idx == 1 then
      wndFirstIcon = wnd
      --self.wndSelectedIcon = wnd
    end
     
    wnd:SetSprite(arStrMacroIcons[idx])
    local strSelectedIconSprite = self.curIconSelection:GetSprite()
    if strSelectedIconSprite == arStrMacroIcons[idx] then
      self:SelectIcon(wnd)
    end
  end
    
  if self.wndSelectedIcon == nil and wndFirstIcon ~= nil then 
    self:SelectIcon(wndFirstIcon) -- select the first icon
  end
    
  self.wndIconList:ArrangeChildrenTiles()  
end

function AMPBuilds:DestroyIconWnd()
  if not ( self.wndIcon == nil ) then
    self.wndIcon:Destroy()
    self.wndIcon = nil
    self.wndSelectedIcon = nil
  end
end

function AMPBuilds:OnIconOK()
    if self.wndSelectedIcon ~= nil then
        -- assign the selected icon to the icon in edit wnd
        local strIcon = self.wndSelectedIcon:GetSprite()
        self.curIconSelection:SetSprite( strIcon )
    end
  self:DestroyIconWnd()
end

function AMPBuilds:OnIconCancel()
  self:DestroyIconWnd()
end

function AMPBuilds:SelectIcon(wnd)
  -- hide old selection frame
  if self.wndSelectedIcon ~= nil then
    local wndSelectedFrame = self.wndSelectedIcon:FindChild("IconSelectedFrame")
    wndSelectedFrame:Show(false)
  end
    
  -- show new selection frame
  self.wndSelectedIcon = wnd
  local wndSelectedFrame = self.wndSelectedIcon:FindChild("IconSelectedFrame")
  wndSelectedFrame:Show(true)
end

function AMPBuilds:OnIconSelect(wndHandler, wndControl)
  self:SelectIcon(wndControl)   
end

function AMPBuilds:OnIconClick()
  self:OpenIconWnd()
end

function AMPBuilds:OnAMPBuildBeginDragDrop(wndHandler, wndControl, x, y, bDragDropStarted )
  if wndHandler ~= wndControl then
    return false
  end

  local ampBuild = wndControl:GetParent():GetParent():GetData()
  local macroId = ampBuild.nMacroId
  local macro = MacrosLib.GetMacro(macroId)
  if not macro then
    self:AMPPrint("The macro associated with this AMP build has been deleted.")
  end

  local strSprite = wndControl:FindChild("Icon"):GetSprite()
  Apollo.BeginDragDrop(wndControl, "DDMacro", strSprite, macroId)
  return true
end

function AMPBuilds:CreateMacro(strSprite_, macroName_)
  local nMacroId = MacrosLib.CreateMacro()
  local tParam = {  nId = nMacroId,
                    bGlobal = false,
                    sName = macroName_,
                    sSprite = strSprite_,
                    sCmds = "/eval CallAMPBuildsMacro(\"" .. macroName_ .. "\")", } 
    
  self:SaveMacro(tParam)
  return nMacroId
end

function AMPBuilds:SaveMacro(tParam)
    MacrosLib.SetMacroData(tParam.nId, false, tParam.sName, tParam.sSprite, tParam.sCmds)
    MacrosLib:SaveMacros()
end

function AMPBuilds:DeleteMacro(nMacroId)
  if nMacroId == nil then
    return
  end  

  MacrosLib.DeleteMacro(nMacroId)
  MacrosLib:SaveMacros()
end

function AMPBuilds:AMPPrint(msg, nChannel, sender)
  if not msg then
    return
  end

  ChatSystemLib.PostOnChannel(nChannel or ChatSystemLib.ChatChannel_System, msg, sender or "AMP Builds")
end

function AMPBuilds:ConfirmAmpSwitch(buildData, nRespecCost)
  if self.dlgConfirm == nil then
    self.dlgConfirm = Apollo.LoadForm(self.xmlDoc, "ConfirmDlg", nil, self)
    self.dlgConfirm:SetAnchorOffsets(self.config.ConfirmDlgPosition.nLeft, self.config.ConfirmDlgPosition.nTop, self.config.ConfirmDlgPosition.nRight, self.config.ConfirmDlgPosition.nBottom)
    self.dlgConfirm:SetData(buildData)
    self.dlgConfirm:FindChild("Icon"):SetSprite(buildData.strSprite)
    self.dlgConfirm:FindChild("Name"):SetText(buildData.strName)
    self.dlgConfirm:FindChild("CashWindow"):SetAmount(nRespecCost)
    self.dlgConfirm:ToFront()
  else
    self.config.ConfirmDlgPosition.nLeft = 20
    self.config.ConfirmDlgPosition.nTop = 20
    self.config.ConfirmDlgPosition.nRight = 338
    self.config.ConfirmDlgPosition.nBottom = 293
    self.dlgConfirm:SetAnchorOffsets(20, 20, 338, 293)
    self.dlgConfirm:ToFront()
  end
end

function AMPBuilds:OnClickRespecOk()
  self:ContinueRespecConfirmed(self.dlgConfirm:GetData())
  self:DestroyConfirmDlg()
end

function AMPBuilds:OnClickRespecCancel()
  self:AMPPrint("AMP Respec cancelled.")
  self:DestroyConfirmDlg()
end

function AMPBuilds:DestroyConfirmDlg()
  if self.dlgConfirm ~= nil then
    self.config.ConfirmDlgPosition.nLeft, self.config.ConfirmDlgPosition.nTop, self.config.ConfirmDlgPosition.nRight, self.config.ConfirmDlgPosition.nBottom = self.dlgConfirm:GetAnchorOffsets()
    self.dlgConfirm:Destroy()
    self.dlgConfirm = nil
  end
end

function AMPBuilds:SwitchBuild(sName)
  -- Preliminary check
  local buildData = self.savedAMPBuildList[sName]
  if buildData == nil then
    self:AMPPrint("Error finding the AMP build associated with this name.")
    return
  end

  local macro = MacrosLib.GetMacro(buildData.nMacroId)
  if not macro then
    self:AMPPrint("The macro associated with this AMP build has been deleted.")
    return
  end

  -- Determine what amps are currently set

  local curAmps = self:GetCurrentEldanAugmentationData()
  local nCount = curAmps.nCount

  local nCountAmpBuild = buildData.nCount
  -- Compare the build to current amps

  --Print("curAmps count: " .. tostring(nCount))
  --Print("AmpBuild count: " .. tostring(nCountAmpBuild))

  local nCountToReset = 0
  local nRespecAmount = 0
  --local tAmpsToReset = {}

  for tier = 3, 1, -1 do
    for idx, amp in pairs(curAmps.tAmps[tier]) do
      --Print("amp: " .. tostring(amp.strTitle))
      local cAmp
      for k, v in pairs(buildData.tUnlockedAugments.tAmps[tier]) do
        if v.nId == amp.nId then
	        cAmp = amp
	      end
      end

      if not cAmp then
        --self:AMPPrint("Id not found: " .. tostring(idx))
        nCountToReset = nCountToReset + 1
        nRespecAmount = nRespecAmount + AbilityBook.GetEldanAugmentationRespecCost(AbilityBook.CodeEnumAMPRespecType.Single, amp.nId)
        --table.insert(tAmpsToReset, amp.nId)
      end
    end
  end

  -- self:AMPPrint("Number of AMPs to reset: " .. tostring(nCountToReset))

  local nMyCash = GameLib.GetPlayerCurrency():GetAmount()
  local bCanAfford = nRespecAmount <= nMyCash
  if not bCanAfford then
    self:AMPPrint("You cannot afford to respec your AMP build!")
    return
  end

  -- Confirm with pop up

  if self.optionsList.optOut == nil or self.optionsList.optOut == false then
    self:ConfirmAmpSwitch(buildData, nRespecAmount)
  else
    self:ContinueRespecConfirmed(buildData)
  end
end

function AMPBuilds:ContinueRespecConfirmed(buildData)
  -- We clicked ok and agreed to continue the respec OR the confirmation was disabled

  if buildData == nil then
    self:AMPPrint("Error finding this AMP build!")
    return
  end

  local macro = MacrosLib.GetMacro(buildData.nMacroId)
  if not macro then
    self:AMPPrint("The macro associated with this AMP build has been deleted.")
    return
  end

  --self:AMPPrint("Switching to AMP Build: " .. buildData.strName)

  -- Determine what amps are currently set

  local curAmps = self:GetCurrentEldanAugmentationData()
  local nCount = curAmps.nCount
  local nCountAmpBuild = buildData.nCount

  -- Compare the build to current amps

  local nCountToReset = 0
  local nRespecAmount = 0
  local tAmpsToReset = { -- list of AMPs that need to be reset
    [1] = { [1] = {}, [2] = {}, [3] = {}, }, -- Tier 1 with 3 rows
    [2] = {}, -- Tier 2
    [3] = {}, -- Tier 3
  }

  for tier = 3, 1, -1 do
    for idx, amp in pairs(curAmps.tAmps[tier]) do
      --Print("amp: " .. tostring(amp.strTitle))
      local cAmp
      for k, v in pairs(buildData.tUnlockedAugments.tAmps[tier]) do
        if v.nId == amp.nId then
	        cAmp = amp
	      end
      end

      if not cAmp then
        --self:AMPPrint("Id not found: " .. tostring(amp.nId))
        nCountToReset = nCountToReset + 1
        nRespecAmount = nRespecAmount + AbilityBook.GetEldanAugmentationRespecCost(AbilityBook.CodeEnumAMPRespecType.Single, amp.nId)
        if tier == 1 then
          -- add to specific tier and row
          table.insert(tAmpsToReset[tier][amp.nDisplayRow], amp.nId)
        else
          -- add to specific tier
          table.insert(tAmpsToReset[tier], amp.nId)
        end
      end
    end
  end

  --self:AMPPrint("Need to respec " .. tostring(nCountToReset) .. " amps.")

  local nMyCash = GameLib.GetPlayerCurrency():GetAmount()
  local bCanAfford = nRespecAmount <= nMyCash
  if not bCanAfford then
    self:AMPPrint("You cannot afford to respec your AMP build!")
    return
  end

  -- We determined we have the cash so begin deactiving amps

  if nCountToReset > 0 then
    self.nAmpRespecUpdates = nCountToReset -- This is how many Update events we have to receive
    self.buildSwitched = true  -- This is set to alert our callback
    -- self.nPlayerCurrency = GameLib.GetPlayerCurrency():GetAmount()  --  Store our beginning cash amount
  end

  local bRespecResult = false
  nRespecAmount = 0

  for tier = 3, 1, -1 do
    if tier == 1 then
      for row = 3, 1, -1 do
        for idx, amp in pairs(tAmpsToReset[tier][row]) do
          --Print("Tier: " .. tostring(tier) .. ", Row: " .. tostring(row) .. ", amp: " .. tostring(amp))

          --self:AMPPrint("k: " .. tostring(idx) .. " - v: " .. tostring(amp))
          local amt = AbilityBook.GetEldanAugmentationRespecCost(AbilityBook.CodeEnumAMPRespecType.Single, amp)

          -- Actual deactivation of amp right here
          bRespecResult = AbilityBook.RespecEldanAugmentations(AbilityBook.CodeEnumAMPRespecType.Single, amp)
          --self:AMPPrint("Respec amp: " .. tostring(amp) .. ", result: " .. tostring(bRespecResult))
          if bRespecResult then
            --self:AMPPrint("Deducting cash: " .. tostring(amt))
            nRespecAmount = nRespecAmount + amt
          else
            self:AMPPrint("Failed to respec!")
            return
          end
        end
      end

    else
      for idx, amp in pairs(tAmpsToReset[tier]) do
        --Print("Tier: " .. tostring(tier) .. ", amp: " .. tostring(amp))

        --self:AMPPrint("k: " .. tostring(idx) .. " - v: " .. tostring(amp))
        local amt = AbilityBook.GetEldanAugmentationRespecCost(AbilityBook.CodeEnumAMPRespecType.Single, amp)

          -- Actual deactivation of amp right here
        bRespecResult = AbilityBook.RespecEldanAugmentations(AbilityBook.CodeEnumAMPRespecType.Single, amp)
        --self:AMPPrint("Respec amp: " .. tostring(amp) .. ", result: " .. tostring(bRespecResult))
        if bRespecResult then
          --self:AMPPrint("Deducting cash: " .. tostring(amt))
          nRespecAmount = nRespecAmount + amt
        else
          self:AMPPrint("Failed to respec!")
          return
        end
      end
    end
  end

  if nRespecAmount > 0 then
    curAmps = self:GetCurrentEldanAugmentationData() -- Refresh our current amps now that we've removed some
  end

  if bRespecResult == ActionSetLib.CodeEnumLimitedActionSetResult.Ok or bRespecResult == true then
    if nRespecAmount > 0 then
      --self:AMPPrint("Respec successful!")
      --self:AMPPrint("AMP Respec cost: " .. tostring(nRespecAmount))
      Sound.Play(Sound.PlayUIVendorSell) --PlayUIVendorBuy)
    end
  elseif bRespecResult == false then
    --self:AMPPrint("Didn't require any amp respecs.")
  else
    self:AMPPrint("Failed to respec. Reason: " .. tostring(bRespecResult))
    return
  end

  if nCountToReset > 0 then
    -- if we reset any AMPs then we must wait for the update event so we get the power points refunded
    --self:AMPPrint("Waiting for Update...") -- now we wait for all the updates to come through
    buildData.nRespecAmount = nRespecAmount
    self.buildData = buildData -- store this build to use in our callback
    return
  else
    buildData.nRespecAmount = 0
  end

  self:CompleteSwitchBuild(buildData)
  self.buildData = nil
end

-- This is our callback function for whenever the character updates their LAS or AMPs
function AMPBuilds:OnCharacterEldanAugmentationsUpdated()
  --Print("CharacterEldanAugmentationsUpdated")

  --local curAmps = self:GetCurrentEldanAugmentationData()
  --Print("Current AMPs set: " .. tostring(curAmps.nCount))

  if self.buildSwitched and self.nAmpRespecUpdates > 0 then
    self.nAmpRespecUpdates = self.nAmpRespecUpdates - 1 -- Subtract one update at a time since each individual amp respec gets its own event fired
  end

  if self.buildSwitched and self.nAmpRespecUpdates == 0 then
    -- Hopefully all went well and we made it here to finally update our amps to the new build we want to switch to
    --Print("CharacterEldanAugmentationsUpdated -- buildSwitched and all updates came through")

    self.buildSwitched = false
    self:CompleteSwitchBuild(self.buildData)
    self.buildData = nil
  end
end

function AMPBuilds:CompleteSwitchBuild(buildData)
  --Print("Finally: CompleteSwitchBuild")

  -- Update cached spec
  
  curAmps = self:GetCurrentEldanAugmentationData() -- Refresh our current amps now that we've removed some
  --Print("AMPs to Activate: " .. tostring(buildData.nCount))

  local nPower = AbilityBook.GetAvailablePower()
  --self:AMPPrint("Starting nPower: " .. tostring(nPower))
  local nSwitched = 0
  local bTryToSave = false

  for tier = 1, 3 do
    for _, amp in pairs(buildData.tUnlockedAugments.tAmps[tier]) do
      local cAmp
      for k, v in pairs(curAmps.tAmps[tier]) do
        if v.nId == amp.nId then
	        cAmp = v
	        break
	      end
      end
      if not cAmp then
        -- This amp wasn't found in our current amp config so we need to Activate it now

        if nPower >= amp.nPowerCost then
          --self:AMPPrint("AMP not currently set: " .. tostring(amp.strTitle))

          -- Cache the change
          bTryToSave = AbilityBook.UpdateEldanAugmentationSpec(amp.nId)
          --Print("bTryToSave: " .. tostring(bTryToSave))
          if bTryToSave ~= ActionSetLib.CodeEnumLimitedActionSetResult.Ok then
            self:AMPPrint("There was a problem trying to set AMP: " .. tostring(amp.strTitle))
            self:AMPPrint("Fatal Error: " .. tostring(bTryToSave))
            --[[
            if GameLib.GetPlayerUnit():IsInCombat() then
              self:HelperCreateMessage(Apollo.GetString("AbilityBuilder_BuildsCantBeChangedCombat"))
            else
              self:HelperCreateMessage(String_GetWeaselString(Apollo.GetString("EldanAugmentation_SaveFailed"), 0))
            end
            ]]
            return
          else
            nPower = nPower - amp.nPowerCost -- AbilityBook.GetAvailablePower()
            --self:AMPPrint("After update nPower: " .. tostring(nPower))
            nSwitched = nSwitched + 1
          end
        else
          self:AMPPrint("Not enough AMP power to Set: " .. amp.strTitle)
        end
      end
    end
  end

  if nSwitched > 0 then
    -- We had to set new AMPs so we need to RequestActionSetChanges for the cached changes to go into effect
    -- Finally Commit cached changes
    --self:AMPPrint("Total AMPs set: " .. tostring(nSwitched))
    local tActionSetIds = ActionSetLib.GetCurrentActionSet()
    local tResultInfo = ActionSetLib.RequestActionSetChanges(tActionSetIds)
    --self:AMPPrint("tResultInfo: " .. tostring(tResultInfo.eResult))
    if tResultInfo.eResult == ActionSetLib.CodeEnumLimitedActionSetResult.RestrictedInPVP or tResultInfo.eResult == ActionSetLib.CodeEnumLimitedActionSetResult.InCombat then
      self:AMPPrint(Apollo.GetString("AbilityBuilder_BuildsCantBeChangedCombat"))
      return
    elseif tResultInfo.eResult ~= ActionSetLib.CodeEnumLimitedActionSetResult.Ok then
      self:AMPPrint(Apollo.GetString("AbilityBuilder_InvalidSetWarning"))

      if tResultInfo.eResult == ActionSetLib.CodeEnumLimitedActionSetResult.MissingTag then
        for idx, strTag in pairs(tResultInfo.tTags) do
          self:AMPPrint(Apollo.GetString("AbilityBuilder_InvalidSetWarning"))
        end
      end
      return
    end
  end

  -- Made it here then all went thru FINALLY 

  local sInfo = string.format("%s - %s", buildData.strName, os.date())

  -- actual respec amount
  -- buildData.nRespecAmount = self.nActualRespecCost -- doesn't seem to get the proper amount from the PlayerCurrencyChanged event
  -- so just use our initial estimated respec cost

  local tStatItem = { strInfo = sInfo, strSprite = buildData.strSprite, nRespecAmt = buildData.nRespecAmount or 0 }
  table.insert(self.statsList, tStatItem)
  self:Redraw()

  -- Play a victory sound because it felt that way to code this thing
  Sound.Play(Sound.PlayUIContractGoldMilestoneAchieved)

  self:AMPPrint("Switch to AMP Build: " .. buildData.strName .. " Complete!")
end

function CallAMPBuildsMacro(macroName_)
  if macroName_ == nil or macroName_ == "" then
    self:AMPPrint("The macro name is empty.")
    return
  end

  AMPBuildsInst:SwitchBuild(macroName_)
end


-- AMPs Display form will draw the amp build

function AMPBuilds:DrawAMPBuild(tData)
  -- To be completed soon.

	-- self.arHighestUnlockedTier = {}
	-- self:OnStartRedrawAll(tData)
end

function AMPBuilds:OnStartRedrawAll(tData)
  if not self.tWndRefs then
		self.tWndRefs = {}
  end
	if self.tWndRefs.wndMain and self.tWndRefs.wndMain:IsValid() then
		self.tWndRefs.wndMain:Destroy()
		self.tWndRefs = {}
	end

	self.tWndRefs.wndMain = Apollo.LoadForm(self.xmlDoc, "AMPsDisplayForm", nil, self)
	self.tWndRefs.wndMiddle = Apollo.LoadForm(self.xmlDoc, "MiddleBG", self.tWndRefs.wndMain:FindChild("ScrollContainer:LightBulbLayer"), self)
	self.tWndRefs.wndMiddle:ToFront()
	self.tWndRefs.wndMessage = self.tWndRefs.wndMain:FindChild("UpdateMessage")
  self.wndAMPsDisplay = self.tWndRefs.wndMain
  self.tWndRefs.wndMain:FindChild("Header"):FindChild("Title"):SetText("AMP Build: " .. tData.strName)

	-- self.tWndRefs.wndMain:FindChild("UpdateMessage"):Show(false, true)
	-- self.tWndRefs.wndMain:FindChild("ResetFrame:ResetBtn"):AttachWindow(self.tWndRefs.wndMain:FindChild("ResetFrame:ResetConfirm"))

	self.nOrigMiddleLeft, self.nOrigMiddleTop, self.nOrigMiddleRight, self.nOrigMiddleBot = self.tWndRefs.wndMiddle:GetAnchorOffsets()
	
	self:RedrawAll(tData)
end

function AMPBuilds:RedrawAll(tData)

--[[

	if not self.tWndRefs.wndMain or not self.tWndRefs.wndMain:IsValid() then
		return
	end

	self.tUnlockedAugments = {}
	local tEldanAugmentationData = tData -- AbilityBook.GetEldanAugmentationData(AbilityBook.GetCurrentSpec())
	if not tEldanAugmentationData then
		return
	end
	
	-- Create a map of dependencies and amps for easy access
	self.tDependencyMap = {}
	self.tAmpMap = {}
	self.tCachedAmps = {}
	for idx, tAmp in pairs(tEldanAugmentationData.tAugments) do
		if tAmp.nEldanAugmentationIdRequired ~= 0 then
			self.tDependencyMap[tAmp.nEldanAugmentationIdRequired] = tAmp
		end
		
		if not self.tAmpMap[tAmp.nCategoryId] then
			self.tAmpMap[tAmp.nCategoryId] = {}
		end
		
		self.tAmpMap[tAmp.nCategoryId][tAmp.nId] = tAmp
	end
	
	-- Categories in tEldanAugmentationData are indexed by the order they were added to the table, not by id.
	self.tCategoryInfo = {}
	for idx, tCategory in pairs(tEldanAugmentationData.tCategories) do
		self.tCategoryInfo[tCategory.nId] = tCategory
	end

	local wndContainer = self.tWndRefs.wndMain:FindChild("ScrollContainer")
	local wndAmps = wndContainer:FindChild("Amps")
	wndAmps:DestroyChildren()

	-- Update Respec Cost
	-- self:OnPlayerCurrencyChanged()

	-- Update Max Size
	local nNumCategories = #karCategoriesInClockwiseOrder
	local nMaxSize = self.nZoomLevel * (19 * 2) -- self.nZoomLevel * (knNumRows * 2)
	local nCenterX = nMaxSize / 2
	local nCenterY = nMaxSize / 2
	wndContainer:FindChild("Amps"):SetAnchorOffsets(0, 0, nMaxSize, nMaxSize)
	wndContainer:FindChild("Triangles"):SetAnchorOffsets(0, 0, nMaxSize, nMaxSize)
	wndContainer:FindChild("LightBulbLayer"):SetAnchorOffsets(0, 0, nMaxSize, nMaxSize)
	wndContainer:RecalculateContentExtents()

	-- Category Triangles (Valid Categories is for Show All)
	local tCategoryPowers = {}
	local tValidCategories = {}

	for idx = 1, #tEldanAugmentationData.tCategories do
		local tCategory = tEldanAugmentationData.tCategories[idx]
		if self.arHighestUnlockedTier[tCategory.nId] and self.arHighestUnlockedTier[tCategory.nId] < tCategory.nHighestTierUnlocked then
			self:HelperCreateMessage(String_GetWeaselString(Apollo.GetString("AMP_TierUnlockedMessage"), tCategory.strName, tCategory.nHighestTierUnlocked), 2)
		end
		self.arHighestUnlockedTier[tCategory.nId] = tCategory.nHighestTierUnlocked

		-- Triangle
		local nTweakValueSize = 9.5
		local nColumn = karCategoriesInClockwiseOrder[tCategory.nId]

		local nLeft = math.cos(2 * math.pi * 4 / nNumCategories)
		local nRight = math.cos(2 * math.pi * 5 / nNumCategories)
		local nRankDepth = math.sin(2 * math.pi * 4 / nNumCategories)

		local nTweakValueGap = nTweakValueSize - 1.5
		local nRankRotateValue = 2 * math.pi * ((nColumn * 2) + 1) / (nNumCategories * 2)
		local nRankCenterX = nCenterX + (self.nZoomLevel * nTweakValueGap * math.cos(nRankRotateValue))
		local nRankCenterY = nCenterY + (self.nZoomLevel * nTweakValueGap * math.sin(nRankRotateValue))

		local nHeight = self.nZoomLevel * nTweakValueSize * math.abs(nRankDepth)
		local nWidth = self.nZoomLevel * nTweakValueSize * math.abs(nRight - nLeft)

		local tTriangle =
		{
			loc =
			{
				nOffsets =
				{
					nRankCenterX - nWidth,
					nRankCenterY - nHeight,
					nRankCenterX + nWidth,
					nRankCenterY + nHeight,
				},
			},
			fRotation = nColumn * 60 + 120,
			strSprite = "",
		}

		-- Triangle Sprite
		local bIsAnInBetweenCategory = tCategory.nId == 4 or tCategory.nId == 5 or tCategory.nId == 6
		if tCategory.nHighestTierUnlocked == 1 then
			tTriangle.strSprite = "spr_AMPS_Triangle_Block2"
		elseif tCategory.nHighestTierUnlocked == 2 then
			tTriangle.strSprite = "spr_AMPS_Triangle_Block1"
		elseif tCategory.nHighestTierUnlocked == 3 then
			tTriangle.strSprite = "spr_AMPS_Triangle_Block0"
		end
		wndContainer:FindChild("Triangles"):AddPixie(tTriangle)

		-- Add to table later for AMPs
		if tCategory.nHighestTierUnlocked > 1 or not bIsAnInBetweenCategory then
			tValidCategories[tCategory.nId] = true -- GOTCHA: Don't use an array as it'll auto shift
		end

		tCategoryPowers[tCategory.nId] = tCategory.fPowerInCategory

		-- Light Bulb
		local wndLightBulb = self.tWndRefs.wndMiddle:FindChild(karCategoryToConstantData[tCategory.nId].strLightBulbName)
		wndLightBulb:SetSprite(tCategory.nHighestTierUnlocked > 1 and karCategoryToConstantData[tCategory.nId].strLightBulbSprite or "")
		wndLightBulb:FindChild("LightBulbText"):SetText("")
	end

	-- Progress Sub Labels
	local wndButtons = self.tWndRefs.wndMain:FindChild("Buttons")
	local wndFloatingLabels = self.tWndRefs.wndMain:FindChild("ScrollContainer"):FindChild("LightBulbLayer")
	for idx = 1, #tEldanAugmentationData.tCategories do
		local strSubLabel = ""
		local tCategory = tEldanAugmentationData.tCategories[idx]
		local bLabelBlackBackground = tCategory.nHighestTierUnlocked == 1 -- and bIsAnInBetweenCategory

		local nProgBarMax = 0
		local nProgBarCurr = 0
		for idx2, tUnlockData in pairs(tCategory.tUnlockedCategories or {}) do
			local strReqCategoryName = karCategoryToConstantData[tUnlockData.nUnlockCategoryId].strName or Apollo.GetString("AMP_NextTier")
			local nReqCategoryPower = tCategoryPowers[tUnlockData.nUnlockCategoryId] -- GOTCHA: tCategory's fPowerInCategory won't be accurate for inbetweens

			if tCategory.nHighestTierUnlocked == 1 and nReqCategoryPower < tUnlockData.nTier2Amount then
				nProgBarCurr = nProgBarCurr + nReqCategoryPower
				nProgBarMax = nProgBarMax + tUnlockData.nTier2Amount
				strSubLabel = String_GetWeaselString(Apollo.GetString("AMP_NextTierProgress"), nReqCategoryPower, tUnlockData.nTier2Amount or "0")
			elseif tCategory.nHighestTierUnlocked == 2 and nReqCategoryPower < tUnlockData.nTier3Amount then
				nProgBarCurr = nProgBarCurr + nReqCategoryPower
				nProgBarMax = nProgBarMax + tUnlockData.nTier3Amount
				strSubLabel = String_GetWeaselString(Apollo.GetString("AMP_NextTierProgress"), nReqCategoryPower, tUnlockData.nTier3Amount or "0")
				strSubLabel = strSubLabel .. "\n" .. Apollo.GetString("AMP_TierTwoExplanation")
			end
		end

		-- TODO: Refactor this code out
		-- Parent Level
		local wndLabel = wndFloatingLabels:FindChild(karCategoryToConstantData[tCategory.nId].strLabelName)
		local bEnableRespec = tCategoryPowers[tCategory.nId] > 0
		local monRespecCost = Money.new()

		monRespecCost:SetAmount(AbilityBook.GetEldanAugmentationRespecCost(AbilityBook.CodeEnumAMPRespecType.Section, tCategory.nId))
		wndLabel:FindChild("LabelProgBar"):SetMax(nProgBarMax)
		wndLabel:FindChild("LabelProgBar"):SetProgress(nProgBarCurr)
		wndLabel:SetTooltip(strSubLabel)
		wndLabel:FindChild("CategoryRespecBtn"):Enable(bEnableRespec)
		wndLabel:FindChild("CategoryRespecBtn"):SetData(tCategory.nId)

		local strTooltip = ""
		
		if bEnableRespec then
			if monRespecCost:GetAmount() > 0 then
				strTooltip = String_GetWeaselString(Apollo.GetString("AMP_ResetCategoryTooltip"), monRespecCost:GetMoneyString())
			else
				strTooltip = Apollo.GetString("Amps_ResetCategoryFree")
			end
		end
		
		wndLabel:FindChild("CategoryRespecBtn"):SetTooltip(strTooltip)

		local wndButtonLabel = self.tWndRefs.wndMain:FindChild("Buttons"):FindChild(karCategoryToConstantData[tCategory.nId].strButtonLabel)
		wndButtonLabel:FindChild("ButtonProgBar"):SetMax(nProgBarMax)
		wndButtonLabel:FindChild("ButtonProgBar"):SetProgress(nProgBarCurr)
		wndButtonLabel:SetTooltip(strSubLabel)
	end

	-- Amps
	local nColumns = knColumnsPerCategory * nNumCategories
	for idx = 1, #tEldanAugmentationData.tAugments do
		local tAmp = tEldanAugmentationData.tAugments[idx]

		if tAmp.eEldanAvailability == AbilityBook.CodeEnumEldanAvailability.Activated then
			if tAmp.bIsCached then
				self.tCachedAmps[tAmp.nId] = tAmp
			end
			
			self.tUnlockedAugments[tAmp.nId] = tAmp.nId
		end

		local wndAmpForm = Apollo.LoadForm(self.xmlDoc, "AmpForm", wndAmps, self)
		wndAmpForm:FindChild("AmpFormButton"):SetData(tAmp)
		wndAmpForm:SetData(tAmp)

		local nHalfWidth = wndAmpForm:GetWidth() / 2
		local nHalfHeight = wndAmpForm:GetHeight() / 2
		if self.nZoomLevel < 20 then
			nHalfWidth = nHalfWidth - 4
			nHalfHeight = nHalfHeight - 4
		elseif self.nZoomLevel < 34 then
			nHalfWidth = nHalfWidth
			nHalfHeight = nHalfHeight
		else
			nHalfWidth = nHalfWidth + 4
			nHalfHeight = nHalfHeight + 4
		end

		local nFudgedDisplayRow = tAmp.nDisplayRow + 4
		local nLineAngle = (tAmp.nDisplayColumn + (karCategoriesInClockwiseOrder[tAmp.nCategoryId] * knColumnsPerCategory)) / nColumns
		local x = nCenterX + (self.nZoomLevel * nFudgedDisplayRow) * math.cos(2 * math.pi * nLineAngle)
		local y = nCenterY + (self.nZoomLevel * nFudgedDisplayRow) * math.sin(2 * math.pi * nLineAngle)
		wndAmpForm:SetAnchorOffsets(x - nHalfWidth, y - nHalfHeight, x + nHalfWidth, y + nHalfHeight)
	end

	-- Connections (TODO REFACTOR)
	for idx, wndAmpForm in pairs(wndAmps:GetChildren()) do
		local tAmp = wndAmpForm:GetData()
		local nEldanAugmentIdRequired = tAmp.nEldanAugmentationIdRequired
		if nEldanAugmentIdRequired and nEldanAugmentIdRequired ~= 0 then

			for idx2, wndPrerequisiteAugment in pairs(wndAmps:GetChildren()) do
				local tPrerequisiteAugment = wndPrerequisiteAugment:GetData()
				if tPrerequisiteAugment.nId == nEldanAugmentIdRequired then
					local nX, nY = wndAmpForm:GetPos()
					local nHalfWidth = wndAmpForm:GetWidth() / 2
					local nHalfHeight = wndAmpForm:GetHeight() / 2
					local nPrerequisiteX, nPrerequisiteY = wndPrerequisiteAugment:GetPos()

					local tAmpRow =
					{
						loc =
						{
							nOffsets =
							{
								nPrerequisiteX + nHalfWidth,
								nPrerequisiteY + nHalfHeight,
								nX + nHalfWidth,
								nY + nHalfHeight,
							},
						},
						bLine = true,
						fWidth = 2,
						cr = ApolloColor.new("9dffffff"),
					}

					wndContainer:FindChild("Amps"):AddPixie(tAmpRow)
				end
			end
		end
	end

	-- Middle Center Piece
	local nMidZoom = self.nZoomLevel / 24
	self.tWndRefs.wndMiddle:SetAnchorOffsets(self.nOrigMiddleLeft * nMidZoom, self.nOrigMiddleTop * nMidZoom, self.nOrigMiddleRight * nMidZoom, self.nOrigMiddleBot * nMidZoom)

	-- Restore scroll pos
	wndContainer:SetVScrollPos(self.nVScrollPos)

	self:RedrawSelections(tEldanAugmentationData, tData)

]]

end

function AMPBuilds:RedrawSelections(tEldanAugmentationData, tData)
	-- AMP icon states
	for idx, wndAmp in pairs(self.tWndRefs.wndMain:FindChild("ScrollContainer:Amps"):GetChildren()) do
		local tAmp = wndAmp:GetData()
		local eEnum = tAmp.eEldanAvailability
		local strSprite = ""

    local ampActive = tData.tUnlockedAugments[tAmp.nId]
    --Print("nId: " .. tAmp.nId .. " ampActive: " .. tostring(ampActive))

    if eEnum == AbilityBook.CodeEnumEldanAvailability.Activated then
      -- Print("Active")
      if ampActive == nil then
        -- Print("Deactivating")
        eEnum = AbilityBook.CodeEnumEldanAvailability.Inactivated
      end
    end

		-- if self.tCachedAmps[tAmp.nId] then
			-- strSprite = "kitBtn_Dropdown_HoloPressedFlyby"	
		-- elseif eEnum == AbilityBook.CodeEnumEldanAvailability.Unavailable then -- 0
		if eEnum == AbilityBook.CodeEnumEldanAvailability.Unavailable then -- 0
			strSprite = knAugmentationUnavailableColor -- (tAmp.nItemIdUnlock and tAmp.nItemIdUnlock ~= 0) and knAugmentationInlaidLockColor
		elseif eEnum == AbilityBook.CodeEnumEldanAvailability.Inaccessible then -- 1
			strSprite = knAugmentationInaccessibleColor
		elseif eEnum == AbilityBook.CodeEnumEldanAvailability.Activated then -- 2
			strSprite = knAugmentationActivatedColor
		else -- Inactivated, some sort of blue		
			-- if self.tUnlockedAugments[tAmp.nId] then
				-- strSprite = knAugmentationActivatedColor
			-- elseif not tAmp.nEldanAugmentationIdRequired or tAmp.nEldanAugmentationIdRequired == 0 or self.tUnlockedAugments[tAmp.nEldanAugmentationIdRequired] then
			if not tAmp.nEldanAugmentationIdRequired or tAmp.nEldanAugmentationIdRequired == 0 or self.tUnlockedAugments[tAmp.nEldanAugmentationIdRequired] then
				strSprite = knAugmentationInactivatedColor
			else
				strSprite = knAugmentationAlmostThereColor
			end
		end

    if ampActive then
      strSprite = knAugmentationActivatedColor
    end

		-- Match Button to sprite (TODO REFACTOR)
		if strSprite == knAugmentationActivatedColor then
			wndAmp:FindChild("AmpFormButton"):ChangeArt("btn_AMPS_HoverOnlyRedCircle")
		elseif strSprite == knAugmentationInactivatedColor then
			wndAmp:FindChild("AmpFormButton"):ChangeArt("btn_AMPS_HoverOnlyBlueCircle")
		else
			wndAmp:FindChild("AmpFormButton"):ChangeArt("")
		end
		wndAmp:SetSprite(strSprite)
	end

	-- Middle Text

	local nTotalPower = AbilityBook.GetTotalPower()
	local nAvailablePower = AbilityBook.GetAvailablePower()
	local wndMiddleBG = self.tWndRefs.wndMain:FindChild("ScrollContainer:LightBulbLayer:MiddleBG")
	local wndAvailablePoints = wndMiddleBG:FindChild("PointsAvailable")
	local wndMaxPoints = wndMiddleBG:FindChild("PointsMaxText")
	local wndProgress = wndMiddleBG:FindChild("PointsProgBar")

	wndAvailablePoints:SetFont(self.nZoomLevel == knLowestZoom and "CRB_HeaderHuge" or "CRB_HeaderGigantic")
	wndAvailablePoints:SetText(nAvailablePower > 0 and nAvailablePower or "")
	wndMaxPoints:SetFont(self.nZoomLevel == knLowestZoom and "CRB_InterfaceSmall" or "CRB_InterfaceMedium_B")
	wndMaxPoints:SetText(nAvailablePower > 0 and "/"..nTotalPower or "") -- String_GetWeaselString(Apollo.GetString("AMP_AvailablePoints"), nAvailablePower)
	wndMiddleBG:FindChild("PointsReady"):Show(nAvailablePower == 0)
	wndProgress:SetMax(nTotalPower)
	wndProgress:SetProgress(nAvailablePower)

end

function AMPBuilds:OnAugmentationTooltip(wndHandler, wndControl, eToolTipType, x, y)
	local tAugment = wndHandler:GetData()
	if not tAugment then
		return
	end

	if self.wndTooltip and self.wndTooltip:IsValid() then
		self.wndTooltip:Destroy()
		self.wndTooltip = nil
	end

		local bAmpPurchased = tAugment.eEldanAvailability == AbilityBook.CodeEnumEldanAvailability.Activated

	local strCategory = karCategoryToConstantData[tAugment.nCategoryId].strName or ""
	self.wndTooltip = wndHandler:LoadTooltipForm("AbilityAMPs.xml", "TooltipForm", self)
	self.wndTooltip:FindChild("NameLabelWindow"):SetText(tAugment.strTitle or "")
	
	local wndPowerCostLabel = self.wndTooltip:FindChild("PowerCostLabelWindow")
	local wndTierLabel = self.wndTooltip:FindChild("TierLabelWindow")
	local wndDescription = self.wndTooltip:FindChild("DescriptionLabelWindow")
	wndPowerCostLabel:SetText(String_GetWeaselString(Apollo.GetString("AMP_PowerCost"), tAugment.nPowerCost or ""))
	wndTierLabel:SetText(String_GetWeaselString(Apollo.GetString("AMP_TierLabel"), strCategory, tAugment.nCategoryTier or ""))
	wndDescription:SetAML("<P TextColor=\"UI_TextHoloBody\" Font=\"CRB_InterfaceSmall\">" .. tAugment.strDescription .. "</P>")
	

	local nTextWidth, nTextHeight = wndDescription:SetHeightToContentHeight()
	
	if bAmpPurchased then
		local wndRespecLabel = self.wndTooltip:FindChild("RespecLabel")
		-- local wndCashWindow = self.wndTooltip:FindChild("RespecCost")
		
		-- wndRespecLabel:Show(true)
		-- wndCashWindow:Show(true)
		-- wndCashWindow:SetAmount(AbilityBook.GetEldanAugmentationRespecCost(AbilityBook.CodeEnumAMPRespecType.Single, tAugment.nId), true)
		
		local nRespecHeight = wndRespecLabel:GetHeight()
		
		nTextHeight = nTextHeight + nRespecHeight
		local nLeft, nTop, nRight, nBottom = wndTierLabel:GetAnchorOffsets()
		wndTierLabel:SetAnchorOffsets(nLeft, nTop - nRespecHeight, nRight, nBottom - nRespecHeight)
		
		local nPowerLeft, nPowerTop, nPowerRight, nPowerBottom = wndPowerCostLabel:GetAnchorOffsets()
		wndPowerCostLabel:SetAnchorOffsets(nPowerLeft, nPowerTop - nRespecHeight, nPowerRight, nPowerBottom - nRespecHeight)
	end
	
	local nTooltipLeft, nTooltipTop, nTooltipRight, nTooltipBottom = unpack(self.wndTooltip:GetOriginalLocation():ToTable().nOffsets)
	self.wndTooltip:SetAnchorOffsets(nTooltipLeft, nTooltipTop, nTooltipRight, nTooltipBottom + nTextHeight)

end

AMPBuildsInst = AMPBuilds:new()
AMPBuildsInst:Init()




function deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end
